#include <windows.h>
#include <stdio.h>
#include "kmp_com.h"
#include "xstdlib.h"
/*
void __cdecl free(void *p)
{
	XFREE(p);
}
void *__cdecl malloc(size_t s)
{
	return XMALLOC(s);
}
*/
void __cdecl operator delete(void *p)
{
	XFREE(p);
}
void *__cdecl operator new(size_t s)
{
	return XMALLOC(s);
}

#include "libsidplay2/include/sidplay/sidplay2.h"

#define KPI_VERSION 0
#define KPI_VERSIONS "t3" /* 1st test release */

struct KMP_CTX_TAG
{
	SidTune *pTune;
	sidplay2 *pPlayer;
	unsigned srate;
	unsigned nch;
	unsigned bits;
	unsigned numsong;
	unsigned uOptimisation;
	unsigned uEnvironment;
	bool bMos6581;
	bool bEndFlag;
};

static const char *sectionname = "sidplay2";
static char iniFilePath[MAX_PATH] = "";

static UINT GetPrivateProfileIntEx(LPCTSTR lpAppName, LPCTSTR lpKeyName, INT nDefault, LPCTSTR lpFileName)
{
	UINT ret;
	TCHAR szBuffer[32];
	ret = GetPrivateProfileInt(lpAppName, lpKeyName, nDefault, lpFileName);
	wsprintf(szBuffer, "%d", ret);
	WritePrivateProfileString(lpAppName, lpKeyName, szBuffer, lpFileName);
	return ret;
}

static bool GetPrivateProfileBoolEx(LPCTSTR lpAppName, LPCTSTR lpKeyName, bool bDefault, LPCTSTR lpFileName)
{
	bool ret;
	TCHAR szBuffer[32];
	if (GetPrivateProfileString(lpAppName, lpKeyName, bDefault ? "True" : "False", szBuffer, 32, lpFileName))
	{
		/* 擪mF */
		ret = szBuffer[0] == 'T' || szBuffer[0] == 't';
	}
	else
		ret = bDefault;
	WritePrivateProfileString(lpAppName, lpKeyName, ret ? "True" : "False", lpFileName);
	return ret;
}


static void Start(KMP_CTX *pctx, unsigned songNum)
{
	pctx->bEndFlag = false;
	if (songNum) pctx->pTune->selectSong(songNum);
	pctx->pPlayer->loadSong(pctx->pTune);
}

static void Setup(KMP_CTX *pctx)
{
	pctx->pTune = 0;
	pctx->pPlayer = 0;
	pctx->nch = GetPrivateProfileBoolEx(sectionname, "Stereo", true, iniFilePath) ? 2 : 1;
	pctx->srate = GetPrivateProfileIntEx(sectionname, "SampleRate", 44100, iniFilePath);
	pctx->bits = GetPrivateProfileBoolEx(sectionname, "ForceOutput8bit", false, iniFilePath) ? 8 : 16;
	pctx->uEnvironment = GetPrivateProfileIntEx(sectionname, "Environment", sid2_envR, iniFilePath);
	if (pctx->uEnvironment > sid2_envR) pctx->uEnvironment = sid2_envR;
	pctx->uOptimisation = GetPrivateProfileIntEx(sectionname, "Optimisation", 0, iniFilePath);
	if (pctx->uOptimisation > SID2_MAX_OPTIMISATION) pctx->uOptimisation = SID2_MAX_OPTIMISATION;
	pctx->bMos6581 = GetPrivateProfileBoolEx(sectionname, "EnableMOS6581", true, iniFilePath);
}

void Close(KMP_CTX *pctx)
{
	if (pctx->pPlayer)
	{
		delete pctx->pPlayer;
		pctx->pPlayer = 0;
	}
	if (pctx->pTune)
	{
		delete pctx->pTune;
		pctx->pTune = 0;
	}
	delete pctx;
}

KMP_CTX *Open(const char *file, unsigned size)
{
	KMP_CTX *pctx = 0;
	do
	{
		pctx = new KMP_CTX;
		if (!pctx) break;

		Setup(pctx);

		pctx->pTune = new SidTune(0);
		if (!pctx->pTune) break;

		if (size)
		{
			if (!pctx->pTune->read((const uint_least8_t*)file, size)) break;
		}
		else
		{
			if (!pctx->pTune->load(file)) break;
		}

		pctx->pPlayer = new sidplay2;
		if (!pctx->pPlayer) break;
		pctx->pPlayer->configure((pctx->nch == 2) ? sid2_stereo : sid2_mono, pctx->srate, pctx->bits, false);
		pctx->pPlayer->environment((sid2_env_t)pctx->uEnvironment);
		pctx->pPlayer->sidModel(pctx->bMos6581 ? SID2_MOS6581 : SID2_MOS8580);
		pctx->pPlayer->optimisation(pctx->uOptimisation);
		pctx->numsong = pctx->pTune->getInfo().songs;
		pctx->pTune->selectSong(0);
		Start(pctx, 0);
		return pctx;
	} while (0);
	if (pctx) Close(pctx);
	return 0;
}

unsigned Write(KMP_CTX *pctx, void *buf, unsigned smp)
{
	unsigned ret;
	if (pctx->bEndFlag) return 0;
	ret = pctx->pPlayer->play(buf, smp);
	if (!ret)
	{
		pctx->bEndFlag = true;
		return 0;
	}
	return ret;
}

unsigned WriteSkip(KMP_CTX *pctx, unsigned smp)
{
	if (pctx->bEndFlag) return 0;
	short tmpbuf[512 * 2];
	unsigned r = smp;
	while (r > sizeof(tmpbuf))
	{
		if (!pctx->pPlayer->play(tmpbuf, sizeof(tmpbuf)))
		{
			pctx->bEndFlag = true;
			return 0;
		}
		r -=  sizeof(tmpbuf);
	}
	if (r && !pctx->pPlayer->play(tmpbuf, r))
	{
		pctx->bEndFlag = true;
		return 0;
	}
	return smp;
}

unsigned SetPosition(KMP_CTX *pctx, unsigned ms)
{
	Start(pctx, ms / 1000);
	return 0;
}

int GetLoopFlag(KMP_CTX *pctx)
{
	return 1;
}

int GetSeekableFlag(KMP_CTX *pctx)
{
	return 1;
}

int GetMultisongFlag(KMP_CTX *pctx)
{
	return 1;
}

unsigned GetFrequency(KMP_CTX *pctx)
{
	return pctx->srate;
}
unsigned GetChannels(KMP_CTX *pctx)
{
	return pctx->nch;
}
unsigned GetSamplebits(KMP_CTX *pctx)
{
	return pctx->bits;
}
unsigned GetLength(KMP_CTX *pctx)
{
	return pctx->numsong * 1000;
}
unsigned GetUnitSamples(KMP_CTX *pctx)
{
	return 44100 * 60 / 1000;	/* 60msec */
}

void XCRTINIT(void)
{
	unsigned pos;
	static const char *iniext = ".ini";
	pos = XDELEXT(iniFilePath, XGETMODULEPATH(), sizeof(iniFilePath));
	XSTRCPYN(iniFilePath + pos, iniext, sizeof(iniFilePath) - pos);

}

const char **GetPluginInfo(void)
{
	static char flags[4] = { 
		KPI_VERSION, 
		1, /* ē\ */
		1, /*  */
		0 /* \ */
	};
	static const char *infos[] = {
		flags,
		"sidplay2.kpi v2.0.7" KPI_VERSIONS " by Mamiya",
		"libsidplay2.dll Copyright (C) 2000 Simon White / reSID Copyright (C) 2001 Dag Lem",
		".sid",".c64",0
	};
	flags[1] = GetPrivateProfileBoolEx(sectionname, "EnableReentrant", true, iniFilePath);
	return infos;
}

